package Nectarine;

use 5.10.0;
use strict;
use warnings;
use Carp;
use POSIX qw/mktime/;
use XML::LibXML;

my $baseurl = "http://www.scenemusic.eu/demovibes/";

=head1 NAME

 Nectarine - A Scenemusic.eu library

=head1 SYNOPSIS

 use Nectarine;
 my $necta = Nectarine->new( { debug => sub { print STDERR join("\n", @_)."\n" } } );
 $necta->updateQueue(1, sub { die "Failed to get info" if ($_[0]); print "Now playing: ".$necta->{'now'}{'song'}." by ".$necta->{'now'}{'artist'}."\n"; });

=head1 DESCRIPTION

Nectarine / scenemusic.eu information library.

=head1 CONSTRUCTOR

=over

=item new ( [ { option1 => bla } ] )

 Instantiates library object with default settings.
 Options and defaults:
	debug		=> sub { carp(@_) },   -- debug function to use, defaults to carp
	fetcher         => local fetcher       -- http fetcher to use. Uses builtin, but can be set to IrssiFetcher::fetch for instance
	verbose		=> 0,                  -- amount of debug info to debug(), higher prints more, 0 only prints errors.

 Returns undef on failure, a necta object otherwise.

 Note that the builtin fetcher is blocking so whenever fetches are needed they will stall.

=back

=cut

sub new
{
	my $class = shift;
	my $o  = shift;
	my $self = {};
	$self = bless($self, $class);
	if (defined($o->{'debug'}))
	{
		return undef unless (ref($o->{'debug'}) eq 'CODE');
		$self->{'debug'}   = $o->{'debug'};
	}
	if (defined($o->{'fetcher'}))
	{
		return undef unless (ref($o->{'fetcher'}) eq 'CODE');
		$self->{'fetcher'} = $o->{'fetcher'};
	}
	$self->{'verbose'} = $o->{'verbose'} // 0;
	$self->_init() or return undef;
	return $self;
}

=head1 METHODS
 
=over
 
=item _debug ( [$level], $line1, [$line2]  )

 Internal function, debugs one or more lines.
 Args:
	level: required debuglevel to print this (set through constructor)
	line1: whatever you want printed. Newline will be added by this function.
 Returns: nothing interesting

=cut

sub _debug
{
	my ($self, $level, @lines) = @_;
	unless(scalar(@lines))
	{
		@lines = ($level);
		$level = 0;
	}
	return if($level > $self->{'verbose'});

	if (defined($self->{'debug'}) && ref($self->{'debug'}) eq 'CODE')
	{
		no strict 'refs';
		foreach my $line (@lines)
		{
			&{$self->{'debug'}}($line);
		}
		return;
	}
	carp(@_);
}

=item _init ( )

 Internal function, initializes library for use. (starts the first queue fetch)
 Args:
 Returns:

=cut

sub _init
{
	my ($self) = @_;
	$self->{'queue'} = [];
}

=item _fetch ( $url, callback, callback_arg, options )

 Internal function, initializes library for use. (starts the first queue fetch)
 Args:
	url         : the url to fetch, i.e. http://www.test.com/lala.html
	callback    : function to call with the result
	callback_arg: argument to pass to result function
	options     : optional hash ref with postdata string or request headers
 Returns:
	0 on success, error string on failure to start fetch

 Note that this function might block, depending on the underlaying fetcher.

=cut

sub _fetch
{
	my ($self, $url, $cb, $cbargs, $opt) = @_;

	if (defined($self->{'fetcher'}) && ref($self->{'fetcher'}) eq 'CODE')
	{
		no strict 'refs';
		$self->_debug(1, "Calling external fetcher for $url");
		return &{$self->{'fetcher'}}($url, $cb, $cbargs, $opt);
	}

	# We'll have to do it ourselves, great!
	# TODO
	return "Sorry, fetcher not yet implemented.";
}

=back
 
=item updateQueue ( forceUpdate, readyCallback )

 Fetches and updates the Nectarine song queue. 
 If no force is specified this function will update the internal queue information and fetch the latest information only if not enough data is present.
 
 Args:
	forceUpdate  : if set to true the latest info will always be fetched online.
	readyCallback: callback that will be called when the queue is done with updating. Only 1 callback may be active at one time. (but that goes for the updater as well)
 Returns: undef on problems (i.e. already updating), 1 on success (note that queuedata won't be there until fetch is done, use the callback if you need the queue)

=cut

sub updateQueue
{
	my ($self, $force, $cb) = @_;
	# TODO, check if necessary
	if (defined($self->{'_queue_update'}))
	{
		return undef;
	}
	$self->{'_queue_update'} = 1;
	$self->{'_queue_updatecb'} = $cb;
	my $res = $self->_fetch($baseurl . "/xml/queue/", \&_newQueueData, $self);
	return 0 if ($res); # Failure
	$self->_updateQueue();
	
	return 1;
}

sub _newQueueNotify
{
	my ($self, $error) = @_;
	if (defined($self->{'_queue_updatecb'}))
	{
		no strict 'refs';
		&{$self->{'_queue_updatecb'}}();
	}
}

sub _newQueueData
{
	my ($content, $headers, $success, $self) = @_;
	delete $self->{'_queue_update'};
	unless($success)
	{
		$self->_debug("Failed to update nectarine queue (download error)....");
		$self->_newQueueNotify("Failure fetching new queue data...");
		return undef;
	}

	# Parse content and add to queue.
	unless(defined($self->{'_parser'}))
	{
		$self->{'_parser'} = new XML::LibXML;
	}
	my $xml;
	local $SIG{'__DIE__'};
	eval
	{
		$xml = $self->{'_parser'}->parse_string($$content);
	};
	if ($@)
	{
		$self->_debug("Failure parsing XML data for queue! - $@");
		$self->_newQueueNotify("Failure parsing new queue data...");
		return undef;
	}
	$self->_debug(1, "Got XML info - parsing....");
	$xml = $xml->getDocumentElement; # Get root node, whatever it's called (playlist)

	# Get now element
	my $now = $xml->getElementsByTagName('now'); # should only by 1 element
	unless(defined($now->[0]))
	{
		$self->_debug("Error in xml... no now element. blegh.");
		$self->_newQueueNotify("Failure parsing new queue data - now element missing... did they change the xml format?");
		return undef;
	}
	$now = $now->[0];
	$self->{'now'} = _parseSongEntry($now->findnodes('entry')->[0]);
	$self->{'now'}{'timeleft'} = $now->findvalue('timeleft');
	$self->{'now'}{'play_time_ux'} = time - $self->{'now'}{'timeleft'};
	$self->{'now'}{'song_done_time_ux'} = $self->{'now'}{'play_time_ux'} + $self->{'now'}{'song_length_s'};
	$self->updateTimeleft(); # make sure we're consistent.

	# Update queue
	$self->{'queue'} = [];
	my $queue = $xml->getElementsByTagName('queue');
	if(defined($queue->[0]))
	{
		foreach my $q ($queue->[0]->findnodes('entry'))
		{
			push @{$self->{'queue'}}, _parseSongEntry($q);
		}
	}

	# And history of course
	$self->{'history'} = [];
	my $history = $xml->getElementsByTagName('history');
	if(defined($history->[0]))
	{
		foreach my $h ($history->[0]->findnodes('entry'))
		{
			push @{$self->{'history'}}, _parseSongEntry($h);
		}
	}

	# Notify a caller if present.
	$self->_newQueueNotify(0);
}

sub _parseSongEntry
{
	# Parse a single song entry, returns hash element
	# Expects a <entry request_time="lala">< .. ></entry> XML element, like what's in <now> or <history> etc.
	my ($xml) = @_;
	my $r;
	$r->{'artist'}       = $xml->findvalue('artist');
	$r->{'artist_id'}    = $xml->findvalue('artist/@id');
	$r->{'song'}         = $xml->findvalue('song');
	$r->{'song_id'}      = $xml->findvalue('song/@id');
	$r->{'song_length'}  = $xml->findvalue('song/@length');
	$r->{'request_time'} = $xml->findvalue('@request_time');
	$r->{'requester'}    = $xml->findvalue('requester');
	$r->{'requester_id'} = $xml->findvalue('requester/@id');
	# Calculate a few possibly useful values
	$r->{'request_time_ux'} = _ymdhms2ux($r->{'request_time'});
	if ($r->{'song_length'} =~ /((\d+):)?(\d+):(\d+)/)
	{
		$r->{'song_length_s'}  = $3*60 + $4;
		# Need plays at time for this, not available in XML (yet).
		# $r->{'song_done_time_ux'} = $r->{'play_time_ux'} + $r->{'song_length_s'};
	}
	return $r;
}

sub _ymdhms2ux
{
	my ($i) = @_;
	return undef unless(defined($i) && $i =~ /(\d{4})-(\d{2})-(\d{2})\s(\d+):(\d+):(\d+)/);
	my $unixtime = mktime ($6, $5, $4, $3, $2 - 1, $1 - 1900);
	return $unixtime;
}

sub _s2hms
{
	my ($i, $full) = @_;
	my $h = int($i / 3600);
	my $m = int(($i % 3600) / 60);
	my $s = (($i % 3600) % 60);
	if ($h || $full)
	{
		return sprintf("%d:%02d:%02d", $h, $m, $s);
	}
	return sprintf("%d:%02d", $m, $s);
}

sub _updateQueue
{
	my ($self) = @_;

	# Check entries to have the correct timestamp.

}

sub updateTimeleft
{
	my ($self) = @_;
	return unless(defined($self->{'now'}) && defined($self->{'now'}{'timeleft'}));
	$self->{'now'}{'timeleft_s'} = $self->{'now'}{'song_done_time_ux'} - time;
	$self->{'now'}{'timeleft_s'} = 0 if ($self->{'now'}{'timeleft_s'} < 0);
	$self->{'now'}{'timeleft_hms'} = _s2hms($self->{'now'}{'timeleft_s'});
	$self->{'now'}{'timeleft'} = $self->{'now'}{'timeleft_hms'};
}

=head1 AUTHOR
  
  BenV / JuNe <benv-necta@junerules.com>
   
=cut

1;
